# Generated by Django 5.1.6 on 2025-02-05 15:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("scheduler", "0019_task_crontask_new_task_id_repeatabletask_new_task_id_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="repeatabletask",
            name="new_task_id",
        ),
        migrations.RemoveField(
            model_name="scheduledtask",
            name="new_task_id",
        ),
        migrations.DeleteModel(
            name="CronTask",
        ),
        migrations.DeleteModel(
            name="RepeatableTask",
        ),
        migrations.DeleteModel(
            name="ScheduledTask",
        ),
    ]
